from typing import Literal

import pandas as pd
import numpy as np
from .errors_exceptions import InvalidArgumentError
from .moving_average import ema, sema, rma


def CCI(
    source: pd.Series,
    length: int = 20,
    constant: float = 0.015,
    method: Literal['sma', 'ema', 'dema', 'tema', 'rma'] = 'sma',
) -> pd.DataFrame:
    """
    Calculate Commodity Channel Index (CCI)  of the input time series
    data.

    Parameters:
    -----------
    source : pd.Series
        The input time series data.
    length : int, optional
        The number of periods to include in the CCI calculation
        (default: 20)
    constant : float, optional
        The constant factor for CCI calculation.
        (default: 0.015)
    method : str, optional
        The method to use for the moving average calculation.
        (default: "sma")
    Returns:
    --------
    pd.DataFrame
        The calculated CCI data as a DataFrame.

    Raises:
    -------
    InvalidArgumentError
        If the method is not 'sma', 'ema', 'sema', or 'rma'.
    """
    source_arr = np.array(source)

    match method:
        case "sma":
            ma = np.convolve(
                source_arr,
                np.ones(length) / length,
                mode="valid"
            )
        case "ema":
            ma = ema(source, length)
        case "dema":
            ma = sema(source, length, 2)
        case "tema":
            ma = sema(source, length, 3)
        case  "rma":
            ma = rma(source, length)
        case _:
            raise InvalidArgumentError(
            "Method must be 'sma', 'ema', 'sema', or 'rma'."
        )

    window = np.lib.stride_tricks.sliding_window_view(
        source_arr,
        length,
    )

    mean_window = np.mean(window, axis=1)
    abs_diff = np.abs(window - mean_window[:, np.newaxis])
    mad = np.mean(abs_diff, axis=1)

    df = pd.DataFrame()
    df["source"] = source[length - 1 :]
    df["mad"] = mad
    df["ma"] = ma
    df["CCI"] = (df["source"] - df["ma"])  / (constant * df["mad"])

    return df



from typing import Literal
import pandas as pd
from .moving_average import sma, ema, sema, rma
from .utils import DynamicTimeWarping

class DidiIndex:
    """
    The DidiIndex class calculates the Didi Index, which is a technical
    indicator used in financial analysis. It calculates the absolute and
    ratio values of the Didi Index, as well as the Dynamic Time Warping
    (DTW) of the absolute value.

    Parameters:
    -----------
    source : pd.Series
        The time series data to calculate the Didi Index for.
    short_length : int
        The number of periods to include in the short moving average
        calculation.
    mid_length : int
        The number of periods to include in the mid moving average
        calculation.
    long_length : int
        The number of periods to include in the long moving average
        calculation.
    method : Literal["sma", "ema", "dema", "tema", "rma"], optional
        The method to use for the moving average calculation.
        (default: "ema")

    Methods:
    --------
    absolute() -> pd.Series:
        Calculate the Didi Index absolute.

    ratio() -> pd.Series:
        Calculate the Didi Index ratio.

    dtw(method:str = 'absolute') -> pd.Series:
        Calculate the Dynamic Time Warping (DTW) of the Didi Index
        absolute.
    """

    def __init__(
        self,
        source: pd.Series,
        short_length: int,
        mid_length: int,
        long_length: int,
        method: Literal["sma", "ema", "dema", "tema", "rma"] = "ema",
    ) -> None:
        """
        Initialize the Didi Index.

        Parameters:
        -----------
        source : pd.Series
            The time series data to calculate the Didi Index for.
        short_length : int
            The number of periods to include in the short moving average
            calculation.
        mid_length : int
            The number of periods to include in the mid moving average
            calculation.
        long_length : int
            The number of periods to include in the long moving average
            calculation.
        method : Literal["sma", "ema", "sema", "rma"]
            The method to use for the moving average calculation.
            (default: "ema")
        """
        self.source = source
        self.short_length = short_length
        self.mid_length = mid_length
        self.long_length = long_length
        self.method = method

        match method:
            case "sma":
                self.short_ma = sma(source, short_length)
                self.mid_ma = sma(source, mid_length)
                self.long_ma = sma(source, long_length)

            case "ema":
                self.short_ma = ema(source, short_length)
                self.mid_ma = ema(source, mid_length)
                self.long_ma = ema(source, long_length)

            case "dema":
                self.short_ma = sema(source, short_length, 2)
                self.mid_ma = sema(source, mid_length, 2)
                self.long_ma = sema(source, long_length, 2)

            case "tema":
                self.short_ma = sema(source, short_length, 3)
                self.mid_ma = sema(source, mid_length, 3)
                self.long_ma = sema(source, long_length, 3)

            case "rma":
                self.short_ma = rma(source, short_length)
                self.mid_ma = rma(source, mid_length)
                self.long_ma = rma(source, long_length)

            case _:
                raise ValueError(
                    "Invalid method provided."
                    "Use 'sma', 'ema', 'dema', 'tema' or 'rma'."
                )

    def absolute(self) -> pd.Series:
        """
        Calculate the Didi Index absolute.

        Returns:
        --------
        pd.Series
            The calculated Didi Index absolute.
        """
        short_didi = self.short_ma - self.mid_ma
        long_didi = self.long_ma - self.mid_ma
        return long_didi - short_didi

    def ratio(self) -> pd.Series:
        """
        Calculate the Didi Index ratio.

        Returns:
        --------
        pd.Series
            The calculated Didi Index ratio.
        """
        short_didi = self.short_ma / self.mid_ma
        long_didi = self.long_ma / self.mid_ma
        return long_didi - short_didi

    def dtw(self, method:Literal['ratio', 'absolute'] = "absolute") -> pd.Series:
        """
        Calculate the Dynamic Time Warping (DTW) of the Didi Index
        absolute.

        Returns:
        --------
        pd.Series
            The calculated DTW of the Didi Index absolute.
        """
        short_didi = (
            DynamicTimeWarping(self.short_ma, self.mid_ma)
            .calculate_dtw_distance(method, True)
        )

        long_didi = (
            DynamicTimeWarping(self.long_ma, self.mid_ma)
            .calculate_dtw_distance(method, True)
        )

        return long_didi - short_didi



import pandas as pd

def stoch(source, high, low, length) -> pd.Series:
    """
    Calculate the Fast Stochastic Oscillator values for the given
    period length.

    Parameters:
    -----------
    length : int
        The length of the stochastic period.

    Returns:
    --------
    pd.Series
        The Fast Stochastic Oscillator values.
    """
    lowest_low = low.rolling(length).min()
    hightest_high = high.rolling(length).max()
    stochastic = (
        100
        * (source - lowest_low)
        / (hightest_high - lowest_low)
    )
    return stochastic.rename("stoch")



from typing import Literal
import pandas as pd
import numpy as np

from .moving_average import sma, ema, sema, rma
from .errors_exceptions import InvalidArgumentError


def TRIX(
    source: pd.Series,
    length: int = 18,
    signal_length: int = 1,
    ma_method: Literal["sma", "ema", "dema", "tema", "rma"] = "ema",
) -> pd.DataFrame:
    """
    Calculate the Triple Exponential Moving Average (TRIX) momentum 
    oscillator indicator.

    Parameters:
    -----------
    source : pd.Series
        The input time series data for calculating TRIX.
    length : int
        The number of periods for the TRIX moving average.
        (default: 18)
    signal_length : int
        The number of periods for the signal line moving average.
        (default: 1)
    ma_method : Literal["sma", "ema", "dema", "tema", "rma"], optional
        The method to use for calculating moving averages, either
        "ema" for Exponential Moving Average or "sma" for Simple Moving
        Average.
        (default: "ema")

    Raises:
    -------
    ValueError
        If an invalid method is provided.
    """
    if isinstance(source, pd.DataFrame):
        raise TypeError("source can't be a DataFrame")

    trix_source = np.log(source)

    match ma_method:
        case "sma":
            trix = (
                sma(sma(sma(trix_source, length), length), length)
                .diff(signal_length)
            ) * 10000

        case "ema":
            trix = (
                ema(ema(ema(trix_source, length), length), length)
                .diff(signal_length)
            ) * 10000

        case "dema":
            trix = (
                sema(sema(sema(trix_source, length, 2), length, 2), length, 2)
                .diff(signal_length)
            ) * 10000

        case "tema":
            trix = (
                sema(sema(sema(trix_source, length, 3), length, 3), length, 3)
                .diff(signal_length)
            ) * 10000

        case "rma":
            trix = (
                rma(rma(rma(trix_source, length), length), length)
                .diff(signal_length)
            ) * 10000

        case _:
            raise InvalidArgumentError(f"'{ma_method}' is not a valid method.")

    return trix.rename("TRIX")



from typing import Literal
import pandas as pd
import numpy as np

def sma(source: pd.Series, length: int) -> pd.Series:
    """
    Calculate the Simple Moving Average (SMA)
    of the input time series data.

    Parameters:
    -----------
    source : pd.Series
        The time series data to calculate the SMA for.
    length : int
        The number of periods to include in the SMA calculation.

    Returns:
    --------
    pd.Series
        The calculated SMA time series data.
    """
    sma_series = source.rolling(length).mean()
    return sma_series.dropna(axis=0)

def ema(source: pd.Series, length: int) -> pd.Series:
    """
    Calculate the Exponential Moving Average (EMA)
    of the input time series data.

    Parameters:
    -----------
    source : pandas.Series
        The time series data to calculate the EMA for.
    length : int
        The number of periods to include in the EMA calculation.

    Returns:
    --------
    pandas.Series
        The calculated EMA time series data.
    """
    sma_series = source.rolling(window=length, min_periods=length).mean()[:length]
    rest = source[length:]
    return (
        pd.concat([sma_series, rest])
        .ewm(span=length, adjust=False)
        .mean()
        .dropna(axis=0)
    )


def sema(source: pd.Series, length: int, smooth: int) -> pd.Series:
    """
    Calculate the Smoothed Exponential Moving Average (SEMA)
    of the input time series data.

    Parameters:
    -----------
    source : pandas.Series
        The time series data to calculate the SEMA for.
    length : int
        The number of periods to include in the SEMA calculation.
    smooth : int
        The smooth of EMAs to calculate.

    Returns:
    --------
    pandas.Series
        The calculeted SEMA time series data.
    """

    emas_dict = {}
    emas_dict["source_1"] = ema(source, length)
    for value in range(2, smooth + 1):
        emas_dict[f"source_{value}"] = ema(
            emas_dict[f"source_{value-1}"],
            length,
        )
    emas_df = pd.DataFrame(emas_dict)
    emas_df["sema"] = (
        emas_df[emas_df.columns[:-1]].diff(axis=1).sum(axis=1) * - 1
        * smooth
        + emas_df[emas_df.columns[-1]]
    )
    sema_series = emas_df["sema"]
    return sema_series.dropna(axis=0)


def _rma_pandas(
    source: pd.Series,
    length: int,
    **kwargs
) -> pd.Series:
    """
    Calculate the Relative Moving Average (RMA) of the input time series
    data.

    Parameters:
    -----------
    source : pandas.Series
        The time series data to calculate the RMA for.
    length : int
        The number of periods to include in the RMA calculation.
    **kwargs : additional keyword arguments
        Additional keyword arguments to pass to the pandas EWM (Exponential
        Weighted Moving Average) function.

    Returns:
    --------
    pandas.Series
        The calculated RMA time series data.

    Note:
    -----
    The first values are different from the TradingView RMA.
    """
    sma_series = (
        source
        .rolling(window=length, min_periods=length)
        .mean()[:length]
    )

    rest = source[length:]

    return (
        pd.concat([sma_series, rest])
        .ewm(alpha=1 / length, **kwargs)
        .mean()
    ).rename("RMA")

def _rma_python(
    source: pd.Series,
    length: int
) -> pd.Series:
    """
    Calculate the Relative Moving Average (RMA) of the input time series
    data using pure python.

    Parameters:
    -----------
    source : pandas.Series
        The time series data to calculate the RMA for.
    length : int
        The number of periods to include in the RMA calculation.

    Returns:
    --------
    pd.Series
        The calculated RMA time series data.

    Note:
    -----
    The pure python version is the only one with precision in the
    initial RMA values. However, with the simple RMA version,
    both pandas and python versions will yield the same precision
    in initial values.
    """
    alpha = 1 / length
    source_pd = _rma_pandas(source, length)[:length]
    source_values = source[length:].to_numpy().tolist()

    rma_series = float(source_pd.dropna().iloc[0])
    rma_list = [rma_series]

    for source_value in source_values:
        rma_series = alpha * source_value + (1 - alpha) * rma_series
        rma_list.append(rma_series)

    rma_series = pd.Series(
        rma_list,
        name="RMA",
        index=source[length - 1:].index
    )

    return rma_series

def rma(
    source: pd.Series,
    length: int,
    method: Literal["numpy", "pandas"] = "numpy"
) -> np.ndarray | pd.Series:
    """
    Calculate the Relative Moving Average (RMA) of the input time series
    data.

    Parameters:
    -----------
    source : pandas.Series
        The time series data to calculate the RMA for.
    length : int
        The number of periods to include in the RMA calculation.
    method : {"numpy", "pandas"}, optional
        The method to use for calculating the RMA, by default "numpy".

    Returns:
    --------
    np.ndarray or pandas.Series
        The calculated RMA time series data.
    """
    match method:
        case "numpy":
            return _rma_python(source, length)
        case "pandas":
            return _rma_pandas(source, length)
        case _:
            raise TypeError("method must be 'numpy' or 'pandas'")



from typing import Literal
import pandas as pd

from .moving_average import sma, ema, sema, rma

from .errors_exceptions import InvalidArgumentError


def tsi(
    source: pd.Series,
    short_length: int = 13,
    long_length: int = 25,
    ma_method: Literal["sma", "ema", "dema", "tema", "rma"] = "ema",
) -> pd.DataFrame:
    """
    Calculate the True Strength Index (TSI) momentum oscillator indicator.

    Parameters:
    -----------
    source : pd.Series
        The input time series data for calculating TSI.
    short_length : int
        The number of periods for the short-term moving average.
        (default: 13)
    long_length : int
        The number of periods for the long-term moving average.
        (default: 25)
    signal_length : int
        The number of periods for the signal line moving average.
        (default: 13)
    ma_method : Literal["sma", "ema", "dema", "tema", "rma"], optional
        The method to use for calculating moving averages, either
        "ema" for Exponential Moving Average or "sma" for Simple Moving
        Average.
        (default: "ema")

    Raises:
    -------
    ValueError
        If an invalid method is provided.
    """
    if isinstance(source, pd.DataFrame):
        raise TypeError("source can't be a DataFrame")

    PC = source.diff().iloc[1:]
    absolute_PC = PC.abs()

    match ma_method:
        case "sma":
            short_smoothed = sma(PC, short_length)
            long_smoothed = sma(short_smoothed, long_length)
            absolute_short_smoothed = sma(absolute_PC, short_length)
            absolute_long_smoothed = sma(absolute_short_smoothed, long_length)

        case "ema":
            short_smoothed = ema(PC, short_length)
            long_smoothed = ema(short_smoothed, long_length)
            absolute_short_smoothed = ema(absolute_PC, short_length)
            absolute_long_smoothed = ema(absolute_short_smoothed, long_length)
        case "dema":
            short_smoothed = sema(PC, short_length, 2)
            long_smoothed = sema(short_smoothed, long_length, 2)
            absolute_short_smoothed = sema(absolute_PC, short_length, 2)
            absolute_long_smoothed = sema(
                absolute_short_smoothed, long_length, 2
            )
        case "tema":
            short_smoothed = sema(PC, short_length, 3)
            long_smoothed = sema(short_smoothed, long_length, 3)
            absolute_short_smoothed = sema(absolute_PC, short_length, 3)
            absolute_long_smoothed = sema(
                absolute_short_smoothed, long_length, 3
            )
        case "rma":
            short_smoothed = rma(PC, short_length)
            long_smoothed = rma(short_smoothed, long_length)
            absolute_short_smoothed = rma(absolute_PC, short_length)
            absolute_long_smoothed = rma(absolute_short_smoothed, long_length)
        case _:
            raise InvalidArgumentError(f"'{ma_method}' is not a valid method.")

    TSI = (long_smoothed / absolute_long_smoothed).rename("TSI")

    return TSI



from typing import Literal
import pandas as pd
from .moving_average import sma, ema, sema, rma
from .utils import DynamicTimeWarping

def MACD(
    source: pd.Series,
    fast_length: int,
    slow_length: int,
    signal_length: int,
    diff_method: Literal["absolute", "ratio", "dtw"] = "absolute",
    ma_method: Literal["sma", "ema", "dema", "tema", "rma"] = "ema",
    signal_method: Literal["sma", "ema", "dema", "tema", "rma"] = "ema",
) -> pd.DataFrame:
    """
    Calculate the Moving Average Convergence Divergence (MACD)
    indicator.

    Parameters:
    -----------
    source : pd.Series
        The input time series data for calculating MACD.
    fast_length : int
        The number of periods for the fast moving average.
    slow_length : int
        The number of periods for the slow moving average.
    signal_length : int
        The number of periods for the signal line moving average.
    method : Literal["ema", "sma"], optional
        The method to use for calculating moving averages, either
        "ema" for Exponential Moving Average or "sma" for Simple Moving
        Average.
        (default: "ema")

    Raises:
    -------
    ValueError
        If an invalid method is provided.
    """
    if isinstance(source, pd.DataFrame):
        raise TypeError("source can't be a DataFrame")

    match ma_method:
        case "sma":
            fast_ma = sma(source, fast_length)
            slow_ma = sma(source, slow_length)
        case "ema":
            fast_ma = ema(source, fast_length)
            slow_ma = ema(source, slow_length)
        case "dema":
            fast_ma = sema(source, fast_length, 2)
            slow_ma = sema(source, slow_length, 2)
        case "tema":
            fast_ma = sema(source, fast_length, 3)
            slow_ma = sema(source, slow_length, 3)
        case "rma":
            fast_ma = rma(source, fast_length)
            slow_ma = rma(source, slow_length)
        case _:
            raise ValueError(f"'{ma_method}' is not a valid method.")
    match diff_method:
        case "absolute":
            macd = (fast_ma - slow_ma).dropna()
        case "ratio":
            macd = (fast_ma / slow_ma).dropna()
        case "dtw":
            macd = (
                DynamicTimeWarping(fast_ma, slow_ma)
                .calculate_dtw_distance("absolute", True)
            )

    match signal_method:
        case "sma":
            macd_signal = sma(macd, signal_length)
        case "ema":
            macd_signal = ema(macd, signal_length)
        case "dema":
            macd_signal = sema(macd, signal_length, 2)
        case "tema":
            macd_signal = sema(macd, signal_length, 3)
        case "rma":
            macd_signal = rma(macd, signal_length)
        case _:
            raise ValueError(f"'{ma_method}' is not a valid method.")

    match diff_method:
        case "absolute":
            histogram = macd - macd_signal
        case "ratio":
            histogram = macd / macd_signal
        case "dtw":
            histogram = (
                DynamicTimeWarping(macd, macd_signal)
                .calculate_dtw_distance("absolute", True)
            )

    macd_df = pd.concat(
        [
            macd.rename('macd'),
            macd_signal.rename("signal"),
            histogram.rename('histogram')
        ], axis=1
    )

    return macd_df



from .moving_average import sma, rma, ema, sema
from .CCI import CCI
from .MACD import MACD
from .RSI import RSI
from .DMI import DMI
from .TRIX import TRIX
from .SMIO import SMIO
from .slow_stoch import slow_stoch
from .stoch import stoch
from .ichimoku import Ichimoku
from .didi_index import didi_index
from .tsi import tsi



from typing import Literal
import pandas as pd
from .stoch import stoch
from .moving_average import sma, ema, sema, rma


def slow_stoch(
    source: pd.Series,
    high: pd.Series,
    low: pd.Series,
    k_length: int = 14,
    k_smoothing: int = 1,
    d_smoothing: int = 3,
    smoothing_method: Literal["sma", "ema", "dema", "tema", "rma"] = "sma",
) -> tuple[pd.Series, pd.Series]:
    """
    Calculate the Slow Stochastic Oscillator values using the
    specified parameters.

    Parameters:
    -----------
    k_length : int, optional
        The length of the %K line period.
        (default: 14)
    k_smoothing : int, optional
        The smoothing period for the %K line.
        (default: 1)
    d_smoothing : int, optional
        The smoothing period for the %D line.
        (default: 3)
    smoothing_method : Literal["sma", "ema"], optional
        The smoothing method to use for calculations, either 'sma'
        (Simple Moving Average) or 'ema'
        (Exponential Moving Average).
        (default: 'sma')

    Returns:
    --------
    tuple[pd.Series, pd.Series]
        A tuple containing the %K line and %D line values.
    """
    stoch_source = stoch(source, high, low, k_length).dropna()

    match smoothing_method:
        case "sma":
            k = sma(stoch_source, k_smoothing)
            d = sma(k, d_smoothing)

        case "ema":
            k = ema(stoch_source, k_smoothing)
            d = ema(k, d_smoothing)

        case "dema":
            k = sema(stoch_source, k_smoothing, 2)
            d = sema(k, d_smoothing, 2)

        case "tema":
            k = sema(stoch_source, k_smoothing, 3)
            d = sema(k, d_smoothing, 3)

        case "rma":
            k = rma(stoch_source, k_smoothing)
            d = rma(k, d_smoothing)

        case _:
            raise ValueError(f"Invalid smoothing method: {smoothing_method}")

    return k.rename("%K"), d.rename("%D")



from typing import Literal
import pandas as pd

from .moving_average import sma, ema, sema, rma
from .tsi import tsi

from .errors_exceptions import InvalidArgumentError


def SMIO(
    source,
    long_length: int = 20,
    short_length: int = 5,
    signal_length: int = 5,
    ma_method: Literal["sma", "ema", "dema", "tema", "rma"] = "ema",
):
    """
    Calculate the SMI Ergodic Oscillator (SMIO) indicator.

    Parameters:
    -----------
    source : pd.Series
        The input time series data for calculating SMIO.
    long_length : int
        The number of periods for the long-term moving average.
        (default: 20)
    short_length : int
        The number of periods for the short-term moving average.
        (default: 5)
    signal_length : int
        The number of periods for the signal line moving average.
        (default: 5)
    ma_method : Literal["sma", "ema", "dema", "tema", "rma"], optional
        The method to use for calculating moving averages, either
        "ema" for Exponential Moving Average or "sma" for Simple Moving
        Average.
        (default: "ema")

    Raises:
    -------
    ValueError
        If an invalid method is provided.

    Returns:
    --------
    pd.Series
        The Stochastic Momentum Index (SMIO) values.
    """
    if isinstance(source, pd.DataFrame):
        raise TypeError("source can't be a DataFrame")

    erg = tsi(source, short_length, long_length)

    match ma_method:
        case "sma":
            sig = sma(erg, signal_length)
        case "ema":
            sig = ema(erg, signal_length)
        case "dema":
            sig = sema(erg, signal_length, 2)
        case "tema":
            sig = sema(erg, signal_length, 3)
        case "rma":
            sig = rma(erg, signal_length)
        case _:
            raise InvalidArgumentError(f"'{ma_method}' is not a valid method.")

    smio = (erg - sig).rename("SMIO")
    return smio



import pandas as pd
import numpy as np
from .moving_average import rma


class DMI:
    """
    Attributes:
    -----------
    close : pd.Series
        The close values from the DataFrame.
    high : pd.Series
        The high prices from the DataFrame.
    low : pd.Series
        The low prices from the DataFrame.
    length : int
        The lookback period for calculating the Stochastic Oscillator.

    """
    def __init__(
        self,
        dataframe: pd.DataFrame,
        close: str = None,
        high: str = None,
        low: str = None,
    ) -> None:
        """
        Initialize the DMI object with the given data and
        parameters.

        Parameters:
        -----------
        dataframe : pd.DataFrame
            The DataFrame containing the close, high, and low data.
        close : str
            The column name in the DataFrame representing the close
            data.
        high : str, optional
            The column name in the DataFrame representing the high
            data. If not provided,
            it will be inferred from common column names.
        low : str, optional
            The column name in the DataFrame representing the low
            data. If not provided,
            it will be inferred from common column names.
        """
        if not isinstance(dataframe, pd.DataFrame):
            raise ValueError("dataframe param must be a DataFrame")

        ohlc_columns_lowercase = ['high', 'low', 'close']
        ohlc_columns_uppercase = ['High', 'Low', 'Close']
        source_columns = [close, high, low]

        is_na_source = all(
            close is None
            for close in source_columns
        )

        is_hlc_columns_lowercase = all(
            column in dataframe.columns
            for column in ohlc_columns_lowercase
        )

        is_ohlc_columns_uppercase = all(
            column in dataframe.columns
            for column in ohlc_columns_uppercase
        )

        columns_not_found = (
            not is_hlc_columns_lowercase
            and not is_ohlc_columns_uppercase
        )

        if is_na_source and columns_not_found:
            raise ValueError("OHLC columns not found in dataframe")

        if close is None:
            if "Close" in dataframe.columns:
                self.close = dataframe["Close"]
            elif "close" in dataframe.columns:
                self.close = dataframe["close"]
            else:
                self.close = None
        else:
            self.close = dataframe[close]

        if high is None:
            if "High" in dataframe.columns:
                self.high = dataframe["High"]
            elif "high" in dataframe.columns:
                self.high = dataframe["high"]
            else:
                self.high = None
        else:
            self.high = dataframe[high]

        if low is None:
            if "Low" in dataframe.columns:
                self.low = dataframe["Low"]
            elif "low" in dataframe.columns:
                self.low = dataframe["low"]
            else:
                self.low = None
        else:
            self.low = dataframe[low]


    def true_range(self) -> pd.Series:
        """
        Calculate the True Range (TR) values for the given data.

        True Range is the greatest of the following three values:
        1. The current high minus the current low.
        2. The absolute value of the current high minus the previous close.
        3. The absolute value of the current low minus the previous close.

        Returns:
        --------
        pd.Series
            The True Range (TR) values.
        """
        true_range = np.maximum(
            self.high - self.low,
            abs(self.high - self.close.shift()),
            abs(self.low - self.close.shift())
        )
        return true_range

    def adx(
        self,
        adx_smoothing: int = 14,
        di_length: int = 14
    ) -> tuple[pd.Series, pd.Series, pd.Series]:
        """
        Calculate the Average Directional Index (ADX) and related
        directional movement values.

        Parameters:
        -----------
        adx_smoothing : int, optional
            The smoothing period for calculating the ADX.
            (default: 14)
        di_length : int, optional
            The length of the directional movement indicator (DI) period.
            (default: 14)

        Returns:
        --------
        tuple[pd.Series, pd.Series, pd.Series]
            A tuple containing the ADX, Positive Directional Movement
            (+DI), and Negative Directional Movement (-DI) values.
        """
        trur = rma(self.true_range().dropna(), di_length)

        up = self.high.diff().dropna()
        down = -self.low.diff().dropna()

        plusDM = up.where((up > down) & (up > 0), 0)
        minusDM = down.where((down > up) & (down > 0), 0)

        plus = 100 * rma(plusDM, di_length) / trur
        minus = 100 * rma(minusDM, di_length) / trur

        sumDM = plus + minus
        subDM = abs(plus - minus)

        adx = 100 * rma(subDM / sumDM.where(sumDM != 0, 1), adx_smoothing)

        plus = plus.rename("DI+")
        minus = minus.rename("DI-")
        adx = adx.rename("ADX")
        return adx, plus, minus

    def di_difference(
        self,
        adx_smoothing=14,
        di_length=14,
    ) -> tuple[pd.Series, pd.Series]:
        """
        Calculate the difference between the Positive Directional
        Movement (+DI) and Negative Directional Movement (-DI),
        and the ratio of +DI to -DI.

        Parameters:
        -----------
        adx_smoothing : int, optional
            The smoothing period for calculating the ADX.
            (default: 14)
        di_length : int, optional
            The length of the directional movement indicator (DI) period.
            (default: 14)

        Returns:
        --------
        tuple[pd.Series, pd.Series]
            A tuple containing the difference between +DI and -DI and
            the ratio of +DI to -DI.
        """
        _, plus, minus = self.adx(adx_smoothing, di_length)
        di_delta = (plus - minus).rename("DI_Delta")
        di_ratio = (plus / minus).rename("DI_Ratio")
        return di_delta, di_ratio



from typing import Literal
import pandas as pd
from .moving_average import sma, ema, sema, rma
from .utils import DynamicTimeWarping

def didi_index(
        source: pd.Series,
        short_length: int,
        mid_length: int,
        long_length: int,
        ma_method: Literal["sma", "ema", "dema", "tema", "rma"] = "ema",
        method: Literal["absolute", "ratio"] = "absolute",
        use_dtw: bool = False,
    ) -> pd.Series:
    """
    Calculate the Didi Index for the given time series data, using the
    specified moving average method and distance calculation method.

    Parameters:
    -----------
    source : pd.Series
        The time series data to calculate the Didi Index for.
    short_length : int
        The number of periods to include in the short moving average
        calculation.
    mid_length : int
        The number of periods to include in the mid moving average
        calculation.
    long_length : int
        The number of periods to include in the long moving average
        calculation.
    ma_method : Literal["sma", "ema", "dema", "tema", "rma"], optional
        The method to use for the moving average calculation.
        (default: "ema")
    method : Literal["absolute", "ratio"], optional
        The method to use for the distances calculation.
        (default: "absolute")
    """
    match ma_method:
        case "sma":
            short_ma = sma(source, short_length)
            mid_ma = sma(source, mid_length)
            long_ma = sma(source, long_length)

        case "ema":
            short_ma = ema(source, short_length)
            mid_ma = ema(source, mid_length)
            long_ma = ema(source, long_length)

        case "dema":
            short_ma = sema(source, short_length, 2)
            mid_ma = sema(source, mid_length, 2)
            long_ma = sema(source, long_length, 2)

        case "tema":
            short_ma = sema(source, short_length, 3)
            mid_ma = sema(source, mid_length, 3)
            long_ma = sema(source, long_length, 3)

        case "rma":
            short_ma = rma(source, short_length)
            mid_ma = rma(source, mid_length)
            long_ma = rma(source, long_length)

        case _:
            raise ValueError(
                "Invalid method provided."
                "Use 'sma', 'ema', 'dema', 'tema' or 'rma'."
            )

    if use_dtw:
        short_didi = (
            DynamicTimeWarping(short_ma, mid_ma)
            .calculate_dtw_distance(method, True)
        )

        long_didi = (
            DynamicTimeWarping(long_ma, mid_ma)
            .calculate_dtw_distance(method, True)
        )

    elif method == "absolute":
        short_didi = short_ma - mid_ma
        long_didi = long_ma - mid_ma

    elif method == "ratio":
        short_didi = short_ma / mid_ma
        long_didi = long_ma / mid_ma
    else:
        raise ValueError("Invalid method provided. Use 'absolute' or 'ratio'.")

    return long_didi - short_didi
